/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * Deluxe Animator .ANM file player
 * Copyright (c) 1998 by Jari Komppa aka Sol/Trauma.
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * This work is based on some other's work, but the source code had no
 * name. After the work I've done, I think the original author was wise
 * enough not to give an address where the mail bomb might arrive.. =)
 *
 * You wouldn't BELIEVE the original source! Talk about obfuscated C..
 *
 * To gain even SOME usability, the source was completely recoded, adding
 * re-entrancy and target device independency on the way.
 *
 * The .ANM format is full of real-mode sillyness, and the original source
 * was ALSO full of realmode-borland-c-sillyness, and the original
 * coders' *goto* sillyness to aid the silly boreland compiler..
 *
 * It was easier to decipher the accompanied "optimized" assembler
 * file than the C source. Really!
 *
 * Thus I'd say that I reverse-engineered the original code and rebuilt
 * everything. There's nothing left of the original code here.
 *
 * It seems even though the .anm specs (which were with the source)
 * say a record may be split between "largepages", this never happens..
 * (man, what a relief). I've tested several .ANM files and I believe
 * this is the case - and since I don't have any .ANM files which have
 * split records, I can't make support for this feature.
 *
 * Few words about the ANM format;
 * It was clearly optimized for fast real-mode use. Thus, it's inefficient
 * whether you look at the size, speed, etc. Fixed at 320*200*256c.
 * Max. number of frames depends on compression ratio! at minimum
 * it's 256, max is (theoretically) 65534. If a frame compresses down
 * to 50%, it will still take as much as a frame that doesn't
 * compress at all! (since the LargePage can't fit another one of
 * frames of that size, and the "frame split between largepages" feature
 * doesn't seem to be in use although it's in the specs..)
 * The compression algo ("run, skip and dump") works pretty well tho.
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>

#pragma pack(0) /* make sure no padding is made in the following structure */

typedef struct lpfileheaderstruct {
 int id;                /* 4 character ID == "LPF " */
 short maxLps;          /* max # largePages allowed. 256 FOR NOW.   */
 short nLps;            /* # largePages in this file. */
 int nRecords;          /* # records in this file.  65534 is current limit + ring */
 short maxRecsPerLp;    /* # records permitted in an lp. 256 FOR NOW.   */
 short lpfTableOffset;  /* Absolute Seek position of lpfTable.  1280 FOR NOW. */
 int contentType;       /* 4 character ID == "ANIM" */
 short width;           /* Width of screen in pixels. */
 short height;          /* Height of screen in pixels. */
 char variant;          /* 0==ANIM. */
 char version;          /* 0==frame rate in 18/sec, 1= 70/sec */
 char hasLastDelta;     /* 1==Last record is a delta from last-to-first frame. */
 char lastDeltaValid;   /* 0==Ignore ring frame. */
 char pixelType;        /* 0==256 color. */
 char CompressionType;  /* 1==(RunSkipDump) Only one used FOR NOW. */
 char otherRecsPerFrm;  /* 0 FOR NOW. */
 char bitmaptype;       /* 1==320x200, 256-color.  Only one implemented so far. */
 char recordTypes[32];  /* Not yet implemented. */
 int nFrames;           /* Number of actual frames in the file, includes ring frame. */
 short framesPerSecond; /* Number of frames to play per second. */
 short pad2[29];        /* 58 bytes of filler to round up to 128 bytes total. */
} lpfileheader;         /* (comments from original source) */

typedef struct lpdescriptorstruct {
 unsigned short baseRecord;  /* Number of first record in this large page. */
 unsigned short nRecords;
         /* Number of records in lp.
          *  bit 15 of "nRecords" == "has continuation from previous lp".
          *  bit 14 of "nRecords" == "final record continues on next lp".
          * (these are not cared about and thankfully they don't seem to occur)
          */
 unsigned short nBytes; /* Total number of bytes of contents, excluding header. */
} lp_descriptor;

typedef struct anmdatastruc { /* ** read only data ** */
  char * anmdata;     /* Pointer to the whole (raw) anm data */
  char * palette;     /* 768 byte, 64-level palette */
  char * framebuffer; /* 320*200 framebuffer */
  void * curlp;       /* Internal use */
  int nextframe;      /* next frame to be rendered */
  int maxframe;       /* max. frame */
  int looped;         /* changes to 1 when looped */
  int fps;            /* FPS from header */
} ANMDATA; /* Used as interface structure. */

#pragma pack() /* Feel free to pad the rest as you like. */


static char * findnextpage(ANMDATA * temp)
{
  /* Find temp->nextframe :) */
  lp_descriptor *LpArray;
  int i=0;
  LpArray=(void*)(temp->anmdata+128+sizeof(lpfileheader)+1024);
  while (((LpArray+i)->baseRecord > temp->nextframe) || (((LpArray+i)->baseRecord + (LpArray+i)->nRecords) <= temp->nextframe))
    i++;
  temp->curlp=(void*)(temp->anmdata+0xb00+(i*0x10000));
  return (temp->anmdata+0x0b00+(i*0x10000)+sizeof(lp_descriptor)+2);
}

static void decodeframe(char * srcP, char * dstP)
{
 /*
  * Decodes one frame with "run, copy and skip" algorithm.
  * we have short and long versions of RLE, copy and skip,
  * plus 'stop' signal. This was fun to reconstruct from assembler.. :)
  */
  int count, color;
  while (1) { /* We're finished when we get terminate signal :) */
    count=*srcP;
    srcP++;
    if (count==0) { /* Short RLE */
      count=*srcP;
      srcP++;
      color=*srcP;
      srcP++;
      memset(dstP,color,count);
      dstP+=count;
    } else
      if (!(count&0x80)) { /* Short copy */
        memcpy(dstP,srcP,count);
        dstP+=count;
        srcP+=count;
      } else { /* long op or short skip */
        count-=0x80;
        if (count>0) { /* short skip */
          dstP+=count;
        } else { /* long op */
          count=*srcP;
          srcP++;
          count+=(*srcP)<<8;
          srcP++;
          if (count==0) /* stop sign */
            return;
          if (!(count&0x8000)) { /* long skip */
            dstP+=count;
          } else {
            count-=0x8000;
            if (!(count&0x4000)) { /* long copy */
              memcpy(dstP,srcP,count);
              dstP+=count;
              srcP+=count;
            } else { /* and finally, long RLE. */
              count-=0x4000;
              color=*srcP;
              srcP++;
              memset(dstP,color,count);
              dstP+=count;
            }
          }
        }
      }
  }
}

void anm_frame(ANMDATA * temp)
{
 /* Render next frame in animation. */
  unsigned short *indexptr;
  int offset=0,i,destframe;
  lp_descriptor *curlp;
  char *ppointer;
  ppointer=findnextpage(temp);
  curlp=temp->curlp;
  indexptr=(unsigned short*)ppointer;
  destframe = temp->nextframe - curlp->baseRecord;

  for(i = 0; i < destframe; i++) {
    offset += indexptr[i];
	}
  ppointer+=curlp->nRecords*2+offset;
	
  if(ppointer[1]) { /* Skip ExtraBytes & align data */
    ppointer += (4 + (((unsigned short *)ppointer)[1] + (((unsigned short *)ppointer)[1]&1)));
  } else {
		ppointer+=4;
  }

  decodeframe(ppointer,temp->framebuffer);

  temp->nextframe++;
  if (temp->nextframe==temp->maxframe) {
    temp->nextframe=1;
    temp->looped=1;
  }
}

void anm_zap(ANMDATA * temp)
{
  free(temp->framebuffer);
  free(temp->palette);
  free(temp->anmdata);
  free(temp);
}

ANMDATA * anm_open(char * filename)
{
  ANMDATA * temp;
  lpfileheader *lpheader;
  FILE * infile;
  int filesize;
  int i;
  infile=fopen(filename,"rb");
  if (infile==NULL) return NULL;
  temp=malloc(sizeof(ANMDATA));
  memset(temp,0,sizeof(ANMDATA));
  fseek(infile,0,SEEK_END);
  filesize = ftell(infile);
  fseek(infile,0,SEEK_SET);
  temp->anmdata=malloc(filesize);
  temp->palette=malloc(768);
  temp->framebuffer=malloc(64000);
  memset(temp->framebuffer,0,64000);
  fread(temp->anmdata,filesize,1,infile);
  fclose(infile);
  lpheader=(void*)temp->anmdata;
  for(i=0; i<256; i++) {
    temp->palette[i*3+2] = *(temp->anmdata+128+sizeof(lpfileheader)+i*4+0)>>2;
    temp->palette[i*3+1] = *(temp->anmdata+128+sizeof(lpfileheader)+i*4+1)>>2;
    temp->palette[i*3]   = *(temp->anmdata+128+sizeof(lpfileheader)+i*4+2)>>2;
  }
  temp->maxframe=lpheader->nRecords;
  temp->nextframe=0;
  temp->looped=0;
  temp->fps=lpheader->framesPerSecond;
  return temp;
}

/*
  Compressed File Library version of anm_open:

ANMDATA * anm_open_cfl(char * filename)
{
  ANMDATA * temp;
  lpfileheader *lpheader;
  int i;
  temp=malloc(sizeof(ANMDATA));
  memset(temp,0,sizeof(ANMDATA));
  temp->anmdata=CFL_GetFile(filename);
  if (temp->anmdata==NULL) {
    free(temp);
    return NULL;
  }
  temp->palette=malloc(768);
  temp->framebuffer=malloc(64000);
  memset(temp->framebuffer,0,64000);
  lpheader=(void*)temp->anmdata;
  for(i=0; i<256; i++) {
    temp->palette[i*3+2] = *(temp->anmdata+128+sizeof(lpfileheader)+i*4+0)>>2;
    temp->palette[i*3+1] = *(temp->anmdata+128+sizeof(lpfileheader)+i*4+1)>>2;
    temp->palette[i*3]   = *(temp->anmdata+128+sizeof(lpfileheader)+i*4+2)>>2;
  }
  temp->maxframe=lpheader->nRecords;
  temp->nextframe=0;
  temp->looped=0;
  temp->fps=lpheader->framesPerSecond;
  return temp;
}
*/
